from nasap_net import Assembly, BindsiteIdConverter
from nasap_net.algorithms.assembly_separation import separate_product_if_possible


def perform_intra_exchange(
        assembly: Assembly,
        metal_bs: str, leaving_bs: str, entering_bs: str,
        ) -> tuple[Assembly, Assembly | None]:
    """Perform an intra-molecular ligand exchange reaction.

    This function returns the resulting assembly(ies) after performing
    a ligand exchange reaction within a single assembly.

    The ligand exchange process can be summarized as follows:
    - The bond between the metal center (metal_bs) and the leaving ligand
      (leaving_bs) in the assembly is broken.
    - A new bond is formed between the metal center (metal_bs) and the
      entering ligand (entering_bs) in the same assembly.

    Parameters
    ----------
    assembly : Assembly
        The assembly.
    metal_bs : str
        The binding site ID of the metal center.
    leaving_bs : str
        The binding site ID of the leaving ligand.
        This binding site should be bonded to metal_bs.
    entering_bs : str
        The binding site ID of the entering ligand.
        This binding site should not be bonded to any other binding site.

    Returns
    -------
    main_assembly : Assembly
        The assembly containing the metal center and the entering ligand.
        The leaving ligand may be part of this assembly or may be
        contained in a separated assembly (leaving_assembly).
    leaving_assembly : Assembly or None
        The separated leaving assembly, if it exists; otherwise None.

    Notes
    -----
    - This function preserves the component IDs in the input assembly;
        no renaming or prefixing is performed.
    """
    assembly = assembly.deepcopy()
    assembly.remove_bond(metal_bs, leaving_bs)
    assembly.add_bond(entering_bs, metal_bs)

    id_converter = BindsiteIdConverter()
    metal_comp, _ = id_converter.global_to_local(metal_bs)
    # Separate the leaving assembly if possible
    assembly, leaving_assem = separate_product_if_possible(
        assembly, metal_comp)
    return assembly, leaving_assem



from nasap_net.models import BindingSite


def test_binding_site():
    site = BindingSite(component_id="M1", site_id="a")
    assert site.component_id == "M1"
    assert site.site_id == "a"


def test_binding_site_ordering():
    site1 = BindingSite(component_id="M1", site_id="a")
    site2 = BindingSite(component_id="M1", site_id="b")
    site3 = BindingSite(component_id="M2", site_id="a")
    assert site1 < site2
    assert site1 < site3
    assert site2 < site3


def test_binding_site_equality():
    site1 = BindingSite(component_id="M1", site_id="a")
    site2 = BindingSite(component_id="M1", site_id="a")
    site3 = BindingSite(component_id="M1", site_id="b")
    assert site1 == site2
    assert site1 != site3



import networkx as nx
from networkx.algorithms.isomorphism import categorical_node_match

from nasap_net import Assembly

__all__ = ['is_roughly_isomorphic']


def is_roughly_isomorphic(assem1: Assembly, assem2: Assembly) -> bool:
    """Check if two assemblies are roughly isomorphic."""
    node_match = categorical_node_match('component_kind', None)
    # QUESTION: Should we use VF2++ instead of VF2?
    return nx.is_isomorphic(
        assem1.rough_g_snapshot, assem2.rough_g_snapshot,
        node_match=node_match)



import os

import pandas as pd

from nasap_net import explore_reactions

from .lib import reactions_to_df, read_file


def explore_reactions_pipeline(
        assemblies_path: os.PathLike | str,
        components_path: os.PathLike | str,
        config_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None:
    id_to_assembly = read_file(assemblies_path)
    config = read_file(config_path)
    components = read_file(components_path)
    
    metal_kind = config['mle_kinds']['metal']
    leaving_kind = config['mle_kinds']['leaving']
    entering_kind = config['mle_kinds']['entering']

    component_kinds = components['component_kinds']

    if verbose:
        print('Enumerating reactions...')
        
    result = explore_reactions(
        id_to_assembly,
        metal_kind=metal_kind,
        leaving_kind=leaving_kind,
        entering_kind=entering_kind,
        component_structures=component_kinds,
        verbose=verbose,
    )

    if verbose:
        print('Reaction enumeration completed.')

    # Output
    _write_output(
        output_path, reactions_to_df(result),
        overwrite=overwrite,
        verbose=verbose,
    )


def _write_output(
        output_path: os.PathLike | str, reaction_df: pd.DataFrame,
        *,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None:
    if not overwrite and os.path.exists(output_path):
        raise FileExistsError(f'Output file "{output_path}" already exists.')
    if verbose:
        print(f'Saving the results to "{output_path}"...')
    dir_ = os.path.dirname(output_path)
    if dir_:
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
    reaction_df.to_csv(output_path, index=False)
    if verbose:
        print(f'Successfully saved to "{output_path}".')



from .assembly_enumeration import enumerate_assemblies_pipeline
from .assembly_list_concatenation import (
    concatenate_assemblies_pipeline,
    concatenate_assemblies_without_isom_checks)
from .bindsite_capping import cap_bindsites_pipeline
from .bondset_enumeration import enum_bond_subsets_pipeline
from .bondset_to_assembly import bondsets_to_assemblies_pipeline
from .duplicate_exclusion import find_unique_assemblies_pipeline
from .reaction_exploration import explore_reactions_pipeline



from collections.abc import Iterable

from nasap_net.isomorphism import convert_assembly_to_igraph
from nasap_net.models import Assembly
from nasap_net.types import ID


class SeparatedIntoMoreThanTwoPartsError(Exception):
    """Exception raised when an assembly is separated into more than two parts."""
    pass


def separate_if_possible(assembly: Assembly, *, metal_comp_id: ID) -> tuple[Assembly, Assembly | None]:
    """Separate the assembly into product and leaving assemblies if possible.

    If the assembly can be separated into two disconnected sub-assemblies,
    this function will return the two sub-assemblies. Otherwise, it will return
    the original assembly and None.

    The one containing the metal component will be the product assembly.
    `metal_comp_id` will be used to identify the metal component.

    Note that the assembly ID will be set to None in the returned assemblies.

    Parameters
    ----------
    *
    assembly : Assembly
        The assembly to be separated.
    metal_comp_id : ID
        The component ID of the metal center. Used to identify the product assembly.

    Returns
    -------
    product_assembly : Assembly
        The assembly containing the metal center.
    leaving_assembly : Assembly or None
        The separated leaving assembly, if it exists; otherwise None.

    Raises
    ------
    SeparatedIntoMoreThanTwoPartsError
        If the assembly is separated into more than two parts.
    """
    conv_res = convert_assembly_to_igraph(assembly)
    g = conv_res.graph

    group_of_each_vertex: list[int] = g.components().membership
    g1_comp = set()
    g2_comp = set()
    for v_index, group in enumerate(group_of_each_vertex):
        if v_index not in conv_res.core_mapping.inv:
            continue
        comp_id = conv_res.core_mapping.inv[v_index]
        match group:
            case 0:
                g1_comp.add(comp_id)
            case 1:
                g2_comp.add(comp_id)
            case _:
                raise SeparatedIntoMoreThanTwoPartsError()
    if not g2_comp:
        return assembly, None
    assem1 = _create_sub_assembly(assembly, g1_comp)
    assem2 = _create_sub_assembly(assembly, g2_comp)

    if metal_comp_id in assem1.components:
        return assem1, assem2
    else:
        assert metal_comp_id in assem2.components
        return assem2, assem1


def _create_sub_assembly(
        assembly: Assembly, comp_ids: Iterable[ID]) -> Assembly:
    """Create a sub-assembly containing only the specified components.

    Assembly ID will be set to None.
    """
    new_components = {
        comp_id: comp
        for comp_id, comp in assembly.components.items()
        if comp_id in comp_ids
    }
    new_bonds = [
        bond for bond in assembly.bonds
        if set(bond.component_ids) <= set(comp_ids)
    ]
    return assembly.copy_with(components=new_components, bonds=new_bonds)



from collections.abc import Iterable, Mapping
from functools import cached_property

from nasap_net.bondset_enumeration import normalize_bondset_under_sym_ops

from .bondsets_validation import validate_bondsets


class BondsetsComparer:
    """Compare bondsets under symmetry operations."""
    def __init__(
            self,
            first_bondsets: Iterable[Iterable[int]],
            second_bondsets: Iterable[Iterable[int]],
            sym_ops: Mapping[str, Mapping[int, int]] | None = None,
            ) -> None:
        first_bondsets = list(first_bondsets)
        second_bondsets = list(second_bondsets)
        validate_bondsets(first_bondsets)
        validate_bondsets(second_bondsets)
        self._first_bondsets = {
            frozenset(bondset) for bondset in first_bondsets}
        self._second_bondsets = {
            frozenset(bondset) for bondset in second_bondsets}
        self.sym_ops = sym_ops

    @property
    def first_bondsets(self) -> set[frozenset[int]]:
        return self._first_bondsets
    
    @property
    def second_bondsets(self) -> set[frozenset[int]]:
        return self._second_bondsets

    @cached_property
    def are_equal(self) -> bool:
        return not self.only_in_first and not self.only_in_second
    
    @cached_property
    def only_in_first(self) -> set[frozenset[int]]:
        return self.first_bondsets - self.first_to_second.keys()
    
    @cached_property
    def only_in_second(self) -> set[frozenset[int]]:
        return self.second_bondsets - self.second_to_first.keys()
    
    @cached_property
    def first_to_second(self) -> dict[frozenset[int], frozenset[int]]:
        return {
            first: self.normalized_to_second[normalized]
            for first, normalized in self.first_to_normalized.items()
            if normalized in self.normalized_to_second.keys()
            }
    
    @cached_property
    def second_to_first(self) -> dict[frozenset[int], frozenset[int]]:
        return {
            second: self.normalized_to_first[normalized]
            for second, normalized in self.second_to_normalized.items()
            if normalized in self.normalized_to_first.keys()
            }
    
    @cached_property
    def first_to_normalized(self) -> dict[frozenset[int], frozenset[int]]:
        return _bondsets_to_normalized(self.first_bondsets, self.sym_ops)
    
    @cached_property
    def second_to_normalized(self) -> dict[frozenset[int], frozenset[int]]:
        return _bondsets_to_normalized(self.second_bondsets, self.sym_ops)
    
    @cached_property
    def normalized_to_first(self) -> dict[frozenset[int], frozenset[int]]:
        # Assert that there are no duplicate normalized bondsets,
        # i.e., there are no duplicate values in self.first_to_normalized.
        unique_normalized = set(self.first_to_normalized.values())
        assert len(unique_normalized) == len(self.first_to_normalized)
        return {v: k for k, v in self.first_to_normalized.items()}
    
    @cached_property
    def normalized_to_second(self) -> dict[frozenset[int], frozenset[int]]:
        # Assert that there are no duplicate normalized bondsets.
        unique_normalized = set(self.second_to_normalized.values())
        assert len(unique_normalized) == len(self.second_to_normalized)
        return {v: k for k, v in self.second_to_normalized.items()}


def _bondsets_to_normalized(
        bondsets: Iterable[Iterable[int]],
        sym_ops: Mapping[str, Mapping[int, int]] | None = None
        ) -> dict[frozenset[int], frozenset[int]]:
    return {
        frozenset(bondset): frozenset(normalize_bondset_under_sym_ops(
            bondset, sym_ops))
        for bondset in bondsets}



from collections.abc import Mapping
from typing import TypeVar

from nasap_net import Assembly, ReactionBase

T_co = TypeVar("T_co", covariant=True)


def embed_assemblies_into_reaction(
        reaction: ReactionBase[T_co],
        id_to_assembly: Mapping[int, Assembly],
) -> T_co:
    """Embed the assemblies into the reaction."""
    return reaction.to_rich_reaction(id_to_assembly)



import re

__all__ = ['validate_name_of_binding_site']


def validate_name_of_binding_site(binding_site: str) -> None:
    """Check if the binding site is valid.

    The binding site should be a string that follows the pattern 
    '[A-Za-z0-9_-]+'.
    'core' is a reserved keyword for the core binding site.
    """
    if not re.fullmatch(r'[A-Za-z0-9_-]+', binding_site):
        raise ValueError(
            'The binding site should follow the pattern [A-Za-z0-9_-]+.')
    if binding_site == 'core':
        raise ValueError(
            '"core" is a reserved keyword for the core binding site.')



from nasap_net.models import Assembly, AuxEdge, BindingSite, Bond, Component
from nasap_net.reaction_exploration_im.lib import group_equivalent_node_combs


def test_single_binding_site():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX2 = Assembly(
        components={'M0': M, 'X0': X, 'X1': X},
        bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)]
    )

    expected = {
        frozenset({(BindingSite('M0', 0),), (BindingSite('M0', 1),)}),
        frozenset({(BindingSite('X0', 0),), (BindingSite('X1', 0),)}),
    }

    result = group_equivalent_node_combs(
        node_combs=[(BindingSite('M0', 0),), (BindingSite('M0', 1),),
                    (BindingSite('X0', 0),), (BindingSite('X1', 0),)],
        assembly=MX2,
    )

    assert result == expected


def test_binding_site_pairs():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX2 = Assembly(
        components={'M0': M, 'X0': X, 'X1': X},
        bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)]
    )

    expected = {
        frozenset({
            (BindingSite('M0', 0), BindingSite('X0', 0)),
            (BindingSite('M0', 1), BindingSite('X1', 0)),
        }),
        frozenset({
            (BindingSite('M0', 0), BindingSite('X1', 0)),
            (BindingSite('M0', 1), BindingSite('X0', 0)),
        }),
    }

    result = group_equivalent_node_combs(
        node_combs=[(BindingSite('M0', 0), BindingSite('X0', 0)),
                    (BindingSite('M0', 0), BindingSite('X1', 0)),
                    (BindingSite('M0', 1), BindingSite('X0', 0)),
                    (BindingSite('M0', 1), BindingSite('X1', 0))],
        assembly=MX2,
    )

    assert result == expected


def test_aux_edges():
    M = Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)]
    )
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    ML2X2_cis = Assembly(
        components={'M0': M, 'L0': L, 'L1': L, 'X0': X, 'X1': X},
        bonds=[
            Bond('M0', 0, 'X0', 0),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0),
        ]
    )

    expected = {
        frozenset({
            (BindingSite('X0', 0), BindingSite('L0', 0)),
            (BindingSite('X1', 0), BindingSite('L1', 0)),
        }),
        frozenset({
            (BindingSite('X0', 0), BindingSite('L1', 0)),
            (BindingSite('X1', 0), BindingSite('L0', 0)),
        }),
    }

    result = group_equivalent_node_combs(
        node_combs=[
            (BindingSite('X0', 0), BindingSite('L0', 0)),  # trans
            (BindingSite('X0', 0), BindingSite('L1', 0)),  # cis
            (BindingSite('X1', 0), BindingSite('L0', 0)),  # cis
            (BindingSite('X1', 0), BindingSite('L1', 0)),  # trans
        ],
        assembly=ML2X2_cis,
    )

    assert result == expected



from nasap_net.models import Assembly, AuxEdge, Bond, Component, LightAssembly
from nasap_net.models.conversion import convert_light_assemblies_to_rich_ones


def test():
    M = Component(kind='M', sites=[0, 1])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    M_square = Component(
        kind='M(sq)', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])

    components = {
        'M': M,
        'L': L,
        'X': X,
        'M(sq)': M_square,
    }

    light_assemblies = {
        # MX2: X0(0)-(0)M0(1)-(0)X1
        'MX2': LightAssembly(
            id_='MX2',
            components={'X0': 'X', 'M0': 'M', 'X1': 'X'},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]),
        'free_X': LightAssembly(components={'X0': 'X'}, bonds=[]),
        # MLX: (0)L0(1)-(0)M0(1)-(0)X0
        'MLX': LightAssembly(
            components={'L0': 'L', 'M0': 'M', 'X0': 'X'},
            bonds=[Bond('L0', 1, 'M0', 0), Bond('M0', 1, 'X0', 0)]),
        'M(sq)X4': LightAssembly(
            components={'M0': 'M(sq)', 'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0),
                   Bond('M0', 2, 'X2', 0), Bond('M0', 3, 'X3', 0)]),
        'M(sq)L2X2': LightAssembly(
            components={'M0': 'M(sq)', 'L0': 'L', 'L1': 'L', 'X0': 'X', 'X1': 'X'},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'L0', 0),
                   Bond('M0', 2, 'X1', 0), Bond('M0', 3, 'L1', 0)]),
    }

    converted = convert_light_assemblies_to_rich_ones(
        light_assemblies, components)

    assert converted == {
        # MX2: X0(0)-(0)M0(1)-(0)X1
        'MX2': Assembly(
            id_='MX2',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]),
        'free_X': Assembly(components={'X0': X}, bonds=[]),
        # MLX: (0)L0(1)-(0)M0(1)-(0)X0
        'MLX': Assembly(
            components={'L0': L, 'M0': M, 'X0': X},
            bonds=[Bond('L0', 1, 'M0', 0), Bond('M0', 1, 'X0', 0)]),
        'M(sq)X4': Assembly(
            components={'M0': M_square, 'X0': X, 'X1': X, 'X2': X, 'X3': X},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0),
                   Bond('M0', 2, 'X2', 0), Bond('M0', 3, 'X3', 0)]),
        'M(sq)L2X2': Assembly(
            components={'M0': M_square, 'L0': L, 'L1': L, 'X0': X, 'X1': X},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'L0', 0),
                   Bond('M0', 2, 'X1', 0), Bond('M0', 3, 'L1', 0)]),
    }



import itertools
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator
from dataclasses import dataclass

from nasap_net.models import Assembly, BindingSite, Bond
from nasap_net.types import ID
from .lib import extract_unique_site_combinations, separate_if_possible
from .models import MLE, MLEKind, Reaction


class ReactionExplorer(ABC):
    def explore(self) -> Iterator[Reaction]:
        mles = self._iter_mles()
        unique_mles = self._get_unique_mles(mles)
        for mle in unique_mles:
            yield self._perform_reaction(mle)

    @abstractmethod
    def _iter_mles(self) -> Iterator[MLE]:
        pass

    @abstractmethod
    def _get_unique_mles(self, mles: Iterable[MLE]) -> Iterator[MLE]:
        pass

    @abstractmethod
    def _perform_reaction(self, mle: MLE) -> Reaction:
        pass


@dataclass(frozen=True)
class IntraReactionExplorer(ReactionExplorer):
    """Class to explore intra-molecular reactions within an assembly.

    Parameters
    ----------
    assembly : Assembly
        The assembly in which intra-molecular reactions are to be explored.
    mle_kind : MLEKind
        The kinds of components involved in the reaction:
            - `mle_kind.metal`: The component kind of the metal binding site.
            - `mle_kind.leaving`: The component kind of the leaving binding site.
            - `mle_kind.entering`: The component kind of the entering binding site.

    Methods
    -------
    explore() -> Iterator[Reaction]
        Explore and yield all possible intra-molecular reactions within the
        assembly based on the specified MLE kind.
    """
    assembly: Assembly
    mle_kind: MLEKind

    def _iter_mles(self) -> Iterator[MLE]:
        """Get all possible MLEs for intra-molecular reactions in an assembly.

        This function generates all MLEs (combinations of metal binding sites,
        leaving binding sites, and entering binding sites) for intra-molecular
        reactions within a given assembly based on the specified component kinds.

        Returned MLEs meet the following conditions:
          - The metal binding site and leaving binding site are connected to each other.
          - The component kind of the metal binding site is `mle_kind.metal`.
          - The component kind of the leaving binding site is `mle_kind.leaving`.
          - The entering binding site is free and has the component kind `mle_kind.entering`.
        """
        ml_pair = _enum_ml_pair(
            self.assembly,
            metal_kind=self.mle_kind.metal, leaving_kind=self.mle_kind.leaving)

        entering_sites = self.assembly.find_sites(
            has_bond=False, component_kind=self.mle_kind.entering)

        for (metal, leaving), entering in itertools.product(
                ml_pair, entering_sites):
            yield MLE(metal, leaving, entering)

    def _get_unique_mles(self, mles: Iterable[MLE]) -> Iterator[MLE]:
        unique_mle_trios = extract_unique_site_combinations(
            [(mle.metal, mle.leaving, mle.entering) for mle in mles],
             self.assembly)
        for unique_mle in unique_mle_trios:
            metal, leaving, entering = unique_mle.site_comb
            yield MLE(
                metal, leaving, entering,
                duplication=unique_mle.duplication)

    def _perform_reaction(self, mle: MLE) -> Reaction:
        raw_product = (
            self.assembly
                .remove_bond(mle.metal, mle.leaving)
                .add_bond(mle.metal, mle.entering)
        )

        product, leaving = separate_if_possible(
            raw_product, metal_comp_id=mle.metal.component_id)

        return Reaction(
            init_assem=self.assembly,
            entering_assem=None,
            product_assem=product,
            leaving_assem=leaving,
            metal_bs=mle.metal,
            leaving_bs=mle.leaving,
            entering_bs=mle.entering,
            duplicate_count=mle.duplication
        )


@dataclass(frozen=True)
class InterReactionExplorer(ReactionExplorer):
    """Class to explore inter-molecular reactions between two assemblies.

    Parameters
    ----------
    init_assembly : Assembly
        The initial assembly. This assembly contains the metal and leaving
        binding sites.
    entering_assembly : Assembly
        The entering assembly. This assembly contains the entering binding site.
    mle_kind : MLEKind
        The kinds of components involved in the reaction:
            - `mle_kind.metal`: The component kind of the metal binding site.
            - `mle_kind.leaving`: The component kind of the leaving binding site.
            - `mle_kind.entering`: The component kind of the entering binding site.

    Methods
    -------
    explore() -> Iterator[Reaction]
        Explore and yield all possible inter-molecular reactions between the
        two assemblies based on the specified MLE kind.
    """
    init_assembly: Assembly
    entering_assembly: Assembly
    mle_kind: MLEKind

    def _iter_mles(self) -> Iterator[MLE]:
        """Get all possible MLEs for inter-molecular reactions between two assemblies.

        This function generates all MLEs (combinations of metal binding sites,
        leaving binding sites, and entering binding sites) for inter-molecular
        reactions between an initial assembly and an entering assembly
        based on the specified component kinds.

        Returned MLEs meet the following conditions:
          - The metal binding site and leaving binding site are connected to each other.
          - The component kind of the metal binding site is `mle_kind.metal`.
          - The component kind of the leaving binding site is `mle_kind.leaving`.
          - The entering binding site is free and has the component kind `mle_kind.entering`.
        """
        ml_pair = _enum_ml_pair(
            self.init_assembly,
            metal_kind=self.mle_kind.metal, leaving_kind=self.mle_kind.leaving)

        entering_sites = self.entering_assembly.find_sites(
            has_bond=False, component_kind=self.mle_kind.entering)

        for (metal, leaving), entering in itertools.product(
                ml_pair, entering_sites):
            yield MLE(metal, leaving, entering)

    def _get_unique_mles(self, mles: Iterable[MLE]) -> Iterator[MLE]:
        mles1, mles2 = itertools.tee(mles)
        unique_ml_pairs = extract_unique_site_combinations(
            [(mle.metal, mle.leaving) for mle in mles1], self.init_assembly)
        unique_entering_sites = extract_unique_site_combinations(
            [(mle.entering,) for mle in mles2], self.entering_assembly)
        for unique_ml, unique_e in itertools.product(
                unique_ml_pairs, unique_entering_sites):
            metal, leaving = unique_ml.site_comb
            (entering,) = unique_e.site_comb
            yield MLE(
                metal, leaving, entering,
                duplication=unique_ml.duplication * unique_e.duplication)

    def _perform_reaction(self, mle: MLE) -> Reaction:
        def init_renaming_func(comp_id: ID) -> ID:
            return f'init_{comp_id}'

        def entering_renaming_func(comp_id: ID) -> ID:
            return f'entering_{comp_id}'

        # Renaming
        renamed_init_assem = _rename_assembly(
            self.init_assembly, init_renaming_func)
        renamed_entering_assem = _rename_assembly(
            self.entering_assembly, entering_renaming_func)
        renamed_mle = MLE(
            metal=BindingSite(
                component_id=init_renaming_func(mle.metal.component_id),
                site_id=mle.metal.site_id),
            leaving=BindingSite(
                component_id=init_renaming_func(mle.leaving.component_id),
                site_id=mle.leaving.site_id),
            entering=BindingSite(
                component_id=entering_renaming_func(mle.entering.component_id),
                site_id=mle.entering.site_id),
            duplication=mle.duplication
        )

        raw_product = (
            _combine_assemblies(renamed_init_assem, renamed_entering_assem)
                .remove_bond(renamed_mle.metal, renamed_mle.leaving)
                .add_bond(renamed_mle.metal, renamed_mle.entering)
        )

        product, leaving = separate_if_possible(
            raw_product, metal_comp_id=renamed_mle.metal.component_id)
        
        # Double the duplication count if both assemblies are the same.
        # This is because the frequency of "A + A" is twice that of "A + B".
        if self.init_assembly == self.entering_assembly:
            dup = mle.duplication * 2
        else:
            dup = mle.duplication

        return Reaction(
            init_assem=self.init_assembly,
            entering_assem=self.entering_assembly,
            product_assem=product,
            leaving_assem=leaving,
            metal_bs=mle.metal,
            leaving_bs=mle.leaving,
            entering_bs=mle.entering,
            duplicate_count=dup
        )


def _enum_ml_pair(
        assem: Assembly, metal_kind: str, leaving_kind: str
        ) -> set[tuple[BindingSite, BindingSite]]:
    ml_pair: set[tuple[BindingSite, BindingSite]] = set()
    for bond in assem.bonds:
        site1, site2 = bond.sites
        kind1 = assem.get_component_kind_of_site(site1)
        kind2 = assem.get_component_kind_of_site(site2)
        if (kind1, kind2) == (metal_kind, leaving_kind):
            ml_pair.add((site1, site2))
        elif (kind1, kind2) == (leaving_kind, metal_kind):
            ml_pair.add((site2, site1))
    return ml_pair


class ComponentIDCollisionError(Exception):
    pass


def _rename_assembly(
        assembly: Assembly, renaming_func: Callable[[ID], ID]) -> Assembly:
    renamed_components = {
        renaming_func(id_): comp
        for id_, comp in assembly.components.items()}
    renamed_bonds = {
        Bond(comp_id1=renaming_func(site1.component_id), site1=site1.site_id,
             comp_id2=renaming_func(site2.component_id), site2=site2.site_id)
        for (site1, site2) in assembly.bonds}
    return Assembly(components=renamed_components, bonds=renamed_bonds)


def _combine_assemblies(
        init_assem: Assembly, entering_assem: Assembly,
        ) -> Assembly:
    if set(init_assem.components) & set(entering_assem.components):
        raise ComponentIDCollisionError(
            "Component ID collision detected between the two assemblies.")

    new_components = (
            dict(init_assem.components) | dict(entering_assem.components))
    new_bonds = init_assem.bonds | entering_assem.bonds
    return Assembly(components=new_components, bonds=new_bonds)



import os
from collections.abc import Hashable, Iterable, Mapping
from typing import TypeVar

from nasap_net import convert_bondset_to_assembly
from nasap_net.pipelines.lib import read_file, write_output

_T = TypeVar('_T', bound=Hashable)



# ============================================================
# Main Process
# ============================================================

def bondsets_to_assemblies_pipeline(
        bondsets_path: os.PathLike | str,
        structure_data_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False
        ) -> None:
    # Input
    id_to_bondset = read_file(bondsets_path, verbose=verbose)
    structure_input = read_file(structure_data_path, verbose=verbose)

    validate_bond_id_to_bindsites(
        structure_input['bonds_and_their_binding_sites'],
        structure_input['components_and_their_kinds'].keys())
    
    # Main process
    if verbose:
        print('Converting bondsets to assemblies...')
    
    id_to_assembly = {
        id_: convert_bondset_to_assembly(
            bondset, structure_input['components_and_their_kinds'],
            structure_input['bonds_and_their_binding_sites'])
        for id_, bondset in id_to_bondset.items()
        }
    if verbose:
        print('Conversion completed.')
    
    # Output
    write_output(
        output_path, id_to_assembly, overwrite=overwrite, verbose=verbose,
        header='Converted assemblies')


# ============================================================
# Input Processing
# ============================================================


def validate_bond_id_to_bindsites(
        bond_id_to_bindsites: Mapping,
        comp_ids: Iterable[str]
        ) -> None:
    for bond_id, (bs1, bs2) in bond_id_to_bindsites.items():
        if not isinstance(bs1, str) or not isinstance(bs2, str):
            raise ValueError(
                'Values in "bonds_and_their_binding_sites" must be '
                'strings.')
        if bs1 == bs2:
            raise ValueError(
                'Values in "bonds_and_their_binding_sites" must be '
                'different.')
        # bs1 and bs2 must be the form of 'comp_id.local_bindsite_id'
        if not bs1.count('.') == 1 or not bs2.count('.') == 1:
            raise ValueError(
                'Values in "bonds_and_their_binding_sites" must be in the '
                'form of "comp_id.local_bindsite_id", e.g., "M1.a".')
        comp_id1, bs_id1 = bs1.split('.')
        comp_id2, bs_id2 = bs2.split('.')
        for comp_id in (comp_id1, comp_id2):
            if comp_id not in comp_ids:
                raise ValueError(
                    f'Unknown component ID "{comp_id}" in '
                    f'"bonds_and_their_binding_sites".')
        if comp_id1 == comp_id2:
            raise ValueError(
                'The two components in a bond must be different.')



import pytest

from nasap_net.models import Assembly, Bond, Component
from nasap_net.reaction_exploration_im.lib import \
    SeparatedIntoMoreThanTwoPartsError, separate_if_possible


@pytest.fixture
def M():
    return Component(kind='M', sites=[0, 1])

@pytest.fixture
def L():
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X():
    return Component(kind='X', sites=[0])


def test_separation(M, L, X):
    # disconnected_MLX: X0(0)-(0)M0(1) + (0)L0(1)
    disconnected_MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0)]
    )
    MX = Assembly({'X0': X, 'M0': M}, [Bond('X0', 0, 'M0', 0)])
    L = Assembly({'L0': L}, [])
    result = separate_if_possible(disconnected_MLX, metal_comp_id='M0')
    assert result == (MX, L)


def test_no_separation(M, L, X):
    # connected_MLX: X0(0)-(0)M0(1)-(0)L0(1)
    connected_MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'L0', 0)]
    )
    MX = Assembly({'X0': X, 'M0': M}, [Bond('X0', 0, 'M0', 0)])
    result = separate_if_possible(connected_MLX, metal_comp_id='M0')
    assert result == (connected_MLX, None)


def test_separation_error(M, L, X):
    # invalid_MLX: (0)M0(1) + (0)X0 + (0)L0(1)
    invalid_MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[]
    )
    with pytest.raises(SeparatedIntoMoreThanTwoPartsError):
        separate_if_possible(invalid_MLX, metal_comp_id='M0')



from nasap_net.models import Assembly, Bond, Component


def test_assembly():
    L = Component(kind="L", sites={"a", "b"})
    M = Component(kind="M", sites={"a", "b"})
    assembly = Assembly(
        components={"L1": L, "M1": M},
        bonds={Bond(comp_id1="L1", site1="a", comp_id2="M1", site2="a")}
        )
    assert assembly.components == {"L1": L, "M1": M}
    assert assembly.bonds == frozenset({
        Bond(comp_id1="L1", site1="a", comp_id2="M1", site2="a")})



from collections.abc import Mapping

from networkx.utils import graphs_equal

from nasap_net import Assembly, Component

__all__ = ['assemblies_equal']


def assemblies_equal(
        assem1: Assembly, assem2: Assembly,
        component_kinds: Mapping[str, Component]
        ) -> bool:
    """Check if two assemblies are equal.

    Equality here means equal as Python objects, not isomorphism of the
    underlying graphs. Component names, binding site names, and the 
    auxiliary edge types must match.

    See Also
    --------
    nasap_net.is_isomorphic
    """
    return graphs_equal(
        assem1.g_snapshot(component_kinds),
        assem2.g_snapshot(component_kinds))



from typing import Literal, TypeAlias

from nasap_net import InterReactionRich, IntraReactionRich

ReactionEmbedded: TypeAlias = (
        IntraReactionRich | InterReactionRich)


def inter_or_intra(
        reaction: ReactionEmbedded) -> Literal["inter", "intra"]:
    if isinstance(reaction, IntraReactionRich):
        return "intra"
    return "inter"
